/*
    
  This file is a part of Shellp, a 'shell helper' library.
  
  Copyright (C) 2011-2012 Jori Liesenborgs
  
  Contact: jori.liesenborgs@gmail.com

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  
  USA

*/

#include "cmdeval.h"
#include "shellinstance.h"
#include "shellcmdstringarg.h"
#include "shellcmdrealarg.h"
#include "iosystem.h"
#include <stdio.h>

namespace shellp
{

CmdEval::CmdEval(const std::string &cmdName, bool showResult, bool store) : ShellCommand(cmdName)
{
	if (store)
	{
		m_pVarName = new ShellCmdStringArg("variable");
		m_pVarName->setDescription("Variable to store the result of the calculation in.");
		addArgument(m_pVarName);
	}
	else
		m_pVarName = 0;

	m_pExpression = new ShellCmdStringArg("expression");
	m_pExpression->setDescription("Expression to evaluate");
	addArgument(m_pExpression);

	m_showResult = showResult;
	if (showResult)
		setDescription("Calculate a mathematical expression or define a variable and shows the result.");
	else
		setDescription("Calculate a mathematical expression or define a variable without showing the result.");
}

CmdEval::~CmdEval()
{
	if (m_pVarName)
		delete m_pVarName;
	delete m_pExpression;
}

bool CmdEval::execute()
{
	ShellInstance *pInst = ShellInstance::getInstance();
	IOSystem *pIOSys = pInst->getIOSystem();

	/*
	mu::Parser *pParser = pInst->getMathParser();
	mu::value_type result = 0;

	try
	{
		pParser->SetExpr(m_pExpression->getValue());
		result = pParser->Eval();
	}
	catch (mu::Parser::exception_type &e)
	{
		setErrorString(std::string("Math parser cannot handle expression:") + e.GetMsg());
		return false;
	}
	*/

	std::string expression = m_pExpression->getValue();
	double value;

	if (!pInst->evaluateExpression(expression, value))
	{
		setErrorString(pInst->getErrorString());
		return false;
	}


	if (m_pVarName)
	{
		std::string varName = m_pVarName->getValue();

		if (!pInst->storeVariable(varName, value))
		{
			setErrorString(std::string("Unable to store result in variable: ") + pInst->getErrorString());
			return false;
		}
	}

	if (m_showResult)
	{
		if (m_pVarName)
			pIOSys->print("Result: %s = %g", m_pVarName->getValue().c_str(), value);
		else
			pIOSys->print("Result: %g", value);
	}

	return true;
}

CmdEvalList::CmdEvalList(const std::string &cmdName) : ShellCommand(cmdName)
{
	setDescription("List the names of constants and variables currently in use.");
}

CmdEvalList::~CmdEvalList()
{
}

bool CmdEvalList::execute()
{
	ShellInstance *pInst = ShellInstance::getInstance();
	std::vector<std::string> constNames;
	std::vector<double> constValues;
	std::vector<std::string> varNames;
	std::vector<double> varValues;

	pInst->getParserNamesAndValues(constNames, constValues, varNames, varValues);

	if (constNames.size() > 0)
		list(true, constNames, constValues);
	if (varNames.size() > 0)
		list(false, varNames, varValues);

	if (constNames.empty() && varNames.empty())
	{
		ShellInstance *pInst = ShellInstance::getInstance();
		IOSystem *pIOSys = pInst->getIOSystem();

		pIOSys->writeOutputLine("No constants or variables currently defined");
	}

	return true;
}

void CmdEvalList::list(bool isConst, const std::vector<std::string> &names, const std::vector<double> &values)
{
	ShellInstance *pInst = ShellInstance::getInstance();
	IOSystem *pIOSys = pInst->getIOSystem();

	std::string spaces;
	int firstColumnLength = 0;

	for (int i = 0 ; i < names.size() ; i++)
	{
		const std::string &name = names[i];

		if (name.length() > firstColumnLength)
			firstColumnLength = name.length();
	}
	firstColumnLength;

	if (isConst)
		pIOSys->writeOutputLine("CONSTANTS:");
	else
		pIOSys->writeOutputLine("VARIABLES:");

	for (int i = 0 ; i < names.size() ; i++)
	{
		const std::string &name = names[i];
		int nameLen = name.length();
		int spaceLen = firstColumnLength-nameLen;

		std::string spaces;
		spaces.resize(spaceLen, ' ');

		pIOSys->print("  %s%s = %g", name.c_str(), spaces.c_str(), values[i]);
	}
	pIOSys->writeOutputLine("");
}

CmdEvalClearVar::CmdEvalClearVar(const std::string &cmdName) : ShellCommand(cmdName)
{
	setDescription("Clear all currently defined variables.");
}

CmdEvalClearVar::~CmdEvalClearVar()
{
}

bool CmdEvalClearVar::execute()
{
	ShellInstance *pInst = ShellInstance::getInstance();

	pInst->clearVariables();
	return true;
}

} // end namespace
