/*
    
  This file is a part of Shellp, a 'shell helper' library.
  
  Copyright (C) 2011-2012 Jori Liesenborgs
  
  Contact: jori.liesenborgs@gmail.com

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  
  USA

*/

#include "qtiosystem.h"
#include "shellmainwin.h"
#include <stdlib.h>
#include <iostream>

namespace shellp
{

QtIOSystem::QtIOSystem(ShellMainWin *pMainWin)
{
	if (!m_waiter.init())
	{
		std::cerr << "Couldn't initialize QtIOSystem::m_waiter, a vital component" << std::endl;
		exit(-1);
	}
	if (m_mutex.Init() != 0)
	{
		std::cerr << "Couldn't initialize QtIOSystem::m_mutex, a vital component" << std::endl;
		exit(-1);
	}
	m_pMainWin = pMainWin;
}

QtIOSystem::~QtIOSystem()
{
}

bool QtIOSystem::readInputLineInternal(std::string &line)
{
	m_pMainWin->waitingForInput();
	m_waiter.waitForSignal();
	m_mutex.Lock();

	if (!m_enteredLines.empty())
	{
		std::string newLine = *(m_enteredLines.begin());
		m_enteredLines.pop_front();

		line = newLine;
	}
	m_mutex.Unlock();
	return true;
}

void QtIOSystem::writeOutputLine(const std::string &line, bool newline)
{
	m_pMainWin->requestWriteOutputLine(line, newline);
}

void QtIOSystem::enterInputLine(const std::string &line)
{
	m_mutex.Lock();
	m_enteredLines.push_back(line);
	m_mutex.Unlock();
	m_waiter.signal();
}

void QtIOSystem::setPercentageFeedback(double pct)
{
	m_pMainWin->requestNewPercentage(pct);
}

bool QtIOSystem::supports2DPlot() const
{
	return true;
}

bool QtIOSystem::supports3DPlot() const
{
#ifdef SHELLP_CONFIG_OPENGL
	return true;
#else
	return false;
#endif // SHELLP_CONFIG_OPENGL
}

bool QtIOSystem::plot2D(int id, const std::string &title,
			    const std::string &xLabel, const std::string &yLabel,
			    const std::vector<double> &xPoints, const std::vector<double> &yPoints,
			    bool forceShow)
{
	if (!m_pMainWin->requestPlot2D(id, title, xLabel, yLabel, xPoints, yPoints, forceShow))
	{
		setErrorString(m_pMainWin->getErrorString());
		return false;
	}
	return true;
}


bool QtIOSystem::plot3D(int id, const std::string &title, 
			  const std::string &xLabel, const std::string &yLabel,
			  const std::string &zLabel,
			  double xOff, double yOff, double xDist, double yDist,
			  double xMultiplier, double yMultiplier, double zMultiplier,
			  const std::vector<double> &data, int numX, int numY, bool forceShow)
{
	if (!m_pMainWin->requestPlot3D(id, title, xLabel, yLabel, zLabel, xOff, yOff, xDist, yDist, xMultiplier, yMultiplier, zMultiplier, data, numX, numY, forceShow))
	{
		setErrorString(m_pMainWin->getErrorString());
		return false;
	}

	return true;
}

void QtIOSystem::writeStatusLine(const std::string &str)
{
	m_pMainWin->requestStatusLine(str);
}

} // end namespace
