/*
    
  This file is a part of Shellp, a 'shell helper' library.
  
  Copyright (C) 2011-2012 Jori Liesenborgs
  
  Contact: jori.liesenborgs@gmail.com

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  
  USA

*/

#include "shell3dplotwindow.h"
#include "qwtplot3d/qwt3d_gridplot.h"
#include "qwtplot3d/qwt3d_scale.h"
#include <QVBoxLayout>
#include <QMenuBar>
#include <QToolBar>
#include <QMenu>
#include <QAction>
#include <iostream>
#include <cmath>

using namespace Qwt3D;

namespace shellp
{

class ScaledLinearScale : public LinearScale
{
public:
	ScaledLinearScale(double offset, double scaleValue, bool isLog = false)
	{
		m_offset = offset;
		m_scaleValue = scaleValue;
		m_log = isLog;
	}

	~ScaledLinearScale()
	{
	}

	QString ticLabel(unsigned int idx) const
	{
		if (idx < majors_p.size())
		{
			double value = 0;

			if (!m_log)
			{
				value = majors_p[idx];

				value *= m_scaleValue;
				value += m_offset;
			}
			else
			{
				value = majors_p[idx];
				value *= m_scaleValue;
				value = std::exp(value)*m_offset;
			}

			return QString::number(value);
		}
		return QString("");
	}

	Scale *clone() const
	{
		return new ScaledLinearScale(*this);
	}
private:
	double m_offset, m_scaleValue;
	bool m_log;
};

class ScaledLogScale : public LogScale
{
public:
	ScaledLogScale(double offset, double scaleValue)
	{
		m_offset = offset;
		m_scaleValue = scaleValue;
	}

	~ScaledLogScale()
	{
	}

	QString ticLabel(unsigned int idx) const
	{
		if (idx < majors_p.size())
		{
			double value = 0;

			value = majors_p[idx];
			value *= m_scaleValue;
			value = std::exp(value)*m_offset;

			return QString::number(value);
		}
		return QString("");
	}

	Scale *clone() const
	{
		return new ScaledLogScale(*this);
	}
private:
	double m_offset, m_scaleValue;
};


Shell3DPlotWindow::Shell3DPlotWindow(QWidget *pParent) : QWidget(pParent)
{
	m_pRequest = 0;
	m_log = false;
	m_plotPixels = false;

	QVBoxLayout *pLayout = new QVBoxLayout(this);

	/* QMenuBar *pBar = new QMenuBar(this);
	pBar->show();
	pBar->setSizePolicy(QSizePolicy::Minimum, QSizePolicy::Minimum);

	QMenu *pMenu = pBar->addMenu("Options");
	QAction *pAction = pMenu->addAction("Toggle linear/log scale");

	QObject::connect(pAction, SIGNAL(triggered()), this, SLOT(toggleLogLinear()));
	*/

	QToolBar *pBar = new QToolBar(this);
	pBar->show();

	QAction *pAction = pBar->addAction("Toggle linear/log scale");
	QObject::connect(pAction, SIGNAL(triggered()), this, SLOT(toggleLogLinear()));

	pAction = pBar->addAction("Toggle physical/pixels");
	QObject::connect(pAction, SIGNAL(triggered()), this, SLOT(togglePixelsOrPhysical()));

	m_pGridPlot = new GridPlot(this);
	m_pGridPlot->show();

	pLayout->addWidget(pBar);
	pLayout->addWidget(m_pGridPlot);
	pLayout->setContentsMargins(0,0,0,0);

	resize(500, 500);

	hide();
}

Shell3DPlotWindow::~Shell3DPlotWindow()
{
	if (m_pRequest)
		delete m_pRequest;
}

void Shell3DPlotWindow::processRequest(Shell3DPlotRequest *pRequest)
{
	if (m_pRequest)
		delete m_pRequest;

	m_pRequest = pRequest;

	if (!m_log)
		plotLinear();
	else
		plotLog();

	/*
	m_pGridPlot->coordinates()->axes[X1].setLabelString(m_pRequest->getXLabel().c_str());
	m_pGridPlot->coordinates()->axes[X2].setLabelString(m_pRequest->getXLabel().c_str());
	m_pGridPlot->coordinates()->axes[X3].setLabelString(m_pRequest->getXLabel().c_str());
	m_pGridPlot->coordinates()->axes[X4].setLabelString(m_pRequest->getXLabel().c_str());
	m_pGridPlot->coordinates()->axes[Y1].setLabelString(m_pRequest->getYLabel().c_str());
	m_pGridPlot->coordinates()->axes[Y2].setLabelString(m_pRequest->getYLabel().c_str());
	m_pGridPlot->coordinates()->axes[Y3].setLabelString(m_pRequest->getYLabel().c_str());
	m_pGridPlot->coordinates()->axes[Y4].setLabelString(m_pRequest->getYLabel().c_str());
	m_pGridPlot->coordinates()->axes[Z1].setLabelString(m_pRequest->getZLabel().c_str());
	m_pGridPlot->coordinates()->axes[Z2].setLabelString(m_pRequest->getZLabel().c_str());
	m_pGridPlot->coordinates()->axes[Z3].setLabelString(m_pRequest->getZLabel().c_str());
	m_pGridPlot->coordinates()->axes[Z4].setLabelString(m_pRequest->getZLabel().c_str());
	*/
	m_pGridPlot->updateData();
	m_pGridPlot->updateGL();

	if (pRequest->getForceShow())
		show();
}

void Shell3DPlotWindow::plotLinear()
{
	QString title(m_pRequest->getTitle().c_str());
	double **pDataOriginal = m_pRequest->getData();
	int w = m_pRequest->getWidth();
	int h = m_pRequest->getHeight();
	double pixelWidth = m_pRequest->getXPixelDistance();
	double pixelHeight = m_pRequest->getYPixelDistance();
	double xStart = m_pRequest->getXOffset();
	double yStart = m_pRequest->getYOffset();
	double xEnd = xStart + (w-1)*pixelWidth;
	double yEnd = yStart + (h-1)*pixelHeight;

	if (m_plotPixels)
	{
		pixelWidth = 1;
		pixelHeight = 1;
		xStart = 0;
		yStart = 0;
		xEnd = w;
		yEnd = h;
	}

	double **pData = new double*[w];
	for (int i = 0 ; i < w ; i++)
	{
		pData[i] = new double[h];
		memcpy(pData[i], pDataOriginal[i], sizeof(double)*h);
	}

	setWindowTitle(QString("3D Plot: ") + title);

	double minz = 1e300;
	double maxz = -1e300;

	for (int x = 0 ; x < w ; x++)
	{
		for (int y = 0 ; y < h ; y++)
		{
			double z = pData[x][y];

			if (z < minz)
				minz = z;
			if (z > maxz)
				maxz = z;
		}
	}

	//std::cout << "maxz = " << maxz << std::endl;
	//std::cout << "minz = " << minz << std::endl;

	if (minz == maxz)
	{
		if (minz != 0)
			maxz = minz + std::abs(minz)*2.0;
		else
			maxz = 1.0;
	}

	for (int x = 0 ; x < w ; x++)
	{
		for (int y = 0 ; y < h ; y++)
		{
			pData[x][y] -= minz;
			pData[x][y] /= (maxz-minz);
		}
	}

	double x = 1;
	double y = 1;
	double xyScale = 1;
	double zScale = (maxz-minz);

	if ((xEnd-xStart) > (yEnd-yStart))
	{
		y = (pixelHeight*(h-1))/(pixelWidth*(w-1));
		xyScale = (xEnd-xStart);
	}
	else
	{
		x = (pixelWidth*(w-1))/(pixelHeight*(h-1));
		xyScale = (yEnd-yStart);
	}

	m_pGridPlot->createDataset(pData, w, h, 0, x, 0, y, false);

	int numAxes = m_pGridPlot->coordinates()->axes.size();

	m_pGridPlot->coordinates()->axes[X1].setScale(new ScaledLinearScale(xStart, xyScale));
	m_pGridPlot->coordinates()->axes[X2].setScale(new ScaledLinearScale(xStart, xyScale));
	m_pGridPlot->coordinates()->axes[X3].setScale(new ScaledLinearScale(xStart, xyScale));
	m_pGridPlot->coordinates()->axes[X4].setScale(new ScaledLinearScale(xStart, xyScale));
	m_pGridPlot->coordinates()->axes[Y1].setScale(new ScaledLinearScale(yStart, xyScale));
	m_pGridPlot->coordinates()->axes[Y2].setScale(new ScaledLinearScale(yStart, xyScale));
	m_pGridPlot->coordinates()->axes[Y3].setScale(new ScaledLinearScale(yStart, xyScale));
	m_pGridPlot->coordinates()->axes[Y4].setScale(new ScaledLinearScale(yStart, xyScale));
	m_pGridPlot->coordinates()->axes[Z1].setScale(new ScaledLinearScale(minz, zScale));
	m_pGridPlot->coordinates()->axes[Z2].setScale(new ScaledLinearScale(minz, zScale));
	m_pGridPlot->coordinates()->axes[Z3].setScale(new ScaledLinearScale(minz, zScale));
	m_pGridPlot->coordinates()->axes[Z4].setScale(new ScaledLinearScale(minz, zScale));

	for (int i = 0 ; i < numAxes ; i++)
	{
		m_pGridPlot->coordinates()->axes[i].setMajors(5);
		m_pGridPlot->coordinates()->axes[i].setMinors(5);
	}

	for (int i = 0 ; i < w ; i++)
		delete [] pData[i];
	delete [] pData;
}

void Shell3DPlotWindow::plotLog()
{
	QString title(m_pRequest->getTitle().c_str());
	double **pDataOriginal = m_pRequest->getData();
	int w = m_pRequest->getWidth();
	int h = m_pRequest->getHeight();
	double pixelWidth = m_pRequest->getXPixelDistance();
	double pixelHeight = m_pRequest->getYPixelDistance();
	double xStart = m_pRequest->getXOffset();
	double yStart = m_pRequest->getYOffset();
	double xEnd = xStart + (w-1)*pixelWidth;
	double yEnd = yStart + (h-1)*pixelHeight;

	if (m_plotPixels)
	{
		pixelWidth = 1;
		pixelHeight = 1;
		xStart = 0;
		yStart = 0;
		xEnd = w;
		yEnd = h;
	}

	double **pData = new double*[w];
	for (int i = 0 ; i < w ; i++)
	{
		pData[i] = new double[h];
		memcpy(pData[i], pDataOriginal[i], sizeof(double)*h);
	}

	setWindowTitle(QString("3D Plot: ") + title);

	double minz = 1e300;
	double maxz = -1e300;

	for (int x = 0 ; x < w ; x++)
	{
		for (int y = 0 ; y < h ; y++)
		{
			double z = pData[x][y];

			if (z < minz)
				minz = z;
			if (z > maxz)
				maxz = z;
		}
	}

	//std::cout << "maxz = " << maxz << std::endl;
	//std::cout << "minz = " << minz << std::endl;

	if (minz == maxz)
	{
		if (minz != 0)
			maxz = minz + std::abs(minz)*2.0;
		else
			maxz = 1.0;
	}

	for (int x = 0 ; x < w ; x++)
	{
		for (int y = 0 ; y < h ; y++)
		{
			pData[x][y] = std::log(pData[x][y]);
			pData[x][y] -= std::log(minz);
			pData[x][y] /= (std::log(maxz)-std::log(minz));
		}
	}

	double x = 1;
	double y = 1;
	double xyScale = 1;
	double zScale = (std::log(maxz)-std::log(minz));

	if ((xEnd-xStart) > (yEnd-yStart))
	{
		y = (pixelHeight*(h-1))/(pixelWidth*(w-1));
		xyScale = (xEnd-xStart);
	}
	else
	{
		x = (pixelWidth*(w-1))/(pixelHeight*(h-1));
		xyScale = (yEnd-yStart);
	}

	m_pGridPlot->createDataset(pData, w, h, 0, x, 0, y, false);

	int numAxes = m_pGridPlot->coordinates()->axes.size();

	m_pGridPlot->coordinates()->axes[X1].setScale(new ScaledLinearScale(xStart, xyScale));
	m_pGridPlot->coordinates()->axes[X2].setScale(new ScaledLinearScale(xStart, xyScale));
	m_pGridPlot->coordinates()->axes[X3].setScale(new ScaledLinearScale(xStart, xyScale));
	m_pGridPlot->coordinates()->axes[X4].setScale(new ScaledLinearScale(xStart, xyScale));
	m_pGridPlot->coordinates()->axes[Y1].setScale(new ScaledLinearScale(yStart, xyScale));
	m_pGridPlot->coordinates()->axes[Y2].setScale(new ScaledLinearScale(yStart, xyScale));
	m_pGridPlot->coordinates()->axes[Y3].setScale(new ScaledLinearScale(yStart, xyScale));
	m_pGridPlot->coordinates()->axes[Y4].setScale(new ScaledLinearScale(yStart, xyScale));
	m_pGridPlot->coordinates()->axes[Z1].setScale(new ScaledLinearScale(minz, zScale, true));
	m_pGridPlot->coordinates()->axes[Z2].setScale(new ScaledLinearScale(minz, zScale, true));
	m_pGridPlot->coordinates()->axes[Z3].setScale(new ScaledLinearScale(minz, zScale, true));
	m_pGridPlot->coordinates()->axes[Z4].setScale(new ScaledLinearScale(minz, zScale, true));

	for (int i = 0 ; i < numAxes ; i++)
	{
		m_pGridPlot->coordinates()->axes[i].setMajors(5);
		m_pGridPlot->coordinates()->axes[i].setMinors(0);
	}

	for (int i = 0 ; i < w ; i++)
		delete [] pData[i];
	delete [] pData;
}

void Shell3DPlotWindow::toggleLogLinear()
{
	m_log = !m_log;

	if (m_log)
		plotLog();
	else
		plotLinear();

	m_pGridPlot->updateData();
	m_pGridPlot->updateGL();
}

void Shell3DPlotWindow::togglePixelsOrPhysical()
{
	m_plotPixels = !m_plotPixels;

	if (m_log)
		plotLog();
	else
		plotLinear();

	m_pGridPlot->updateData();
	m_pGridPlot->updateGL();
}

} // end namespace
