/*
    
  This file is a part of Shellp, a 'shell helper' library.
  
  Copyright (C) 2011-2012 Jori Liesenborgs
  
  Contact: jori.liesenborgs@gmail.com

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  
  USA

*/

#include "shellmainwin.h"
#include "shell2dplotwindow.h"
#include "shell3dplotwindow.h"
#include "shellinstance.h"
#include "qtiosystem.h"
#include "historylineedit.h"
#include <QVBoxLayout>
#include <QHBoxLayout>
#include <QTextEdit>
#include <QLabel>
#include <QCloseEvent>
#include <QMenuBar>
#include <QMenu>
#include <QAction>
#include <QMessageBox>
#include <QProgressBar>
#include <QStatusBar>

#include <iostream>
#include <cmath>

namespace shellp
{

ShellMainWinThread::ShellMainWinThread(const std::string &shellPrefix, QObject *pParent) : QThread(pParent)
{
	m_shellPrefix = shellPrefix;
}

ShellMainWinThread::~ShellMainWinThread()
{
}

void ShellMainLoop(const std::string &shellPrefix, const std::string &prefix, bool exitOnError);

void ShellMainWinThread::run()
{
	ShellInstance *pInst = ShellInstance::getInstance();
	IOSystem *pIOSys = pInst->getIOSystem();

	std::string prefix = "";
	bool exitOnError = false;

	ShellMainLoop(m_shellPrefix, prefix, exitOnError);
}

ShellMainWin::ShellMainWin(const std::string &title, const std::string &shellPrefix)
{
	resize(800, 600);
	setWindowTitle(title.c_str());
	show();

	QWidget *pCentralWidget = new QWidget(this);
	setCentralWidget(pCentralWidget);

	m_pTextEdit = new QTextEdit(pCentralWidget);
	QLabel *pLabel = new QLabel(shellPrefix.c_str(), pCentralWidget);
	m_pLineEdit = new HistoryLineEdit(pCentralWidget);
	m_pPctBar = new QProgressBar(pCentralWidget);

	QVBoxLayout *pVLayout = new QVBoxLayout(pCentralWidget);
	QHBoxLayout *pHLayout = new QHBoxLayout();
	pVLayout->addWidget(m_pTextEdit);
	pVLayout->addItem(pHLayout);
	pHLayout->setSpacing(5);
	pHLayout->addWidget(pLabel);
	pHLayout->addWidget(m_pLineEdit);
	pHLayout->addWidget(m_pPctBar);


	m_pTextEdit->setReadOnly(true);
	m_pTextEdit->setFocusPolicy(Qt::NoFocus);
	m_pTextEdit->setFont(QFont("Courier"));
	m_pLineEdit->setFont(QFont("Courier"));
	pLabel->setFont(QFont("Courier"));

	m_pPctBar->setMaximumWidth(100);
	m_pPctBar->setRange(0, 10000);
	m_pPctBar->setValue(10000);

	m_pThread = new ShellMainWinThread(shellPrefix, this);

	QMenuBar *pMenuBar = menuBar();
	QMenu *pMenu = pMenuBar->addMenu("Options");
	QAction *pInterruptAction = pMenu->addAction("Interrupt calculation");
	QAction *pForceQuitAction = pMenu->addAction("Force quit");

	QObject::connect(m_pLineEdit, SIGNAL(returnPressed()), this, SLOT(lineEntered()));
	QObject::connect(m_pThread, SIGNAL(finished()), this, SLOT(close()));
	QObject::connect(this, SIGNAL(signalWriteOutputLine(const QString &, bool)),
			 this, SLOT(writeOutputLine(const QString &, bool)));
	QObject::connect(this, SIGNAL(signalWaitingForInput()),
			 this, SLOT(slotWaitingForInput()));
	QObject::connect(pInterruptAction, SIGNAL(triggered()), this, SLOT(slotInterrupt()));
	QObject::connect(pForceQuitAction, SIGNAL(triggered()), this, SLOT(slotForceQuit()));
	QObject::connect(this, SIGNAL(signalProgressBarValue(int)), m_pPctBar, SLOT(setValue(int)));
	QObject::connect(this, SIGNAL(signalNew2DPlotRequest()), SLOT(slotNew2DPlotRequest()));
	QObject::connect(this, SIGNAL(signalNew3DPlotRequest()), SLOT(slotNew3DPlotRequest()));
	QObject::connect(this, SIGNAL(signalStatusLine(const QString &)), this, SLOT(slotStatusLine(const QString &)));

	if (m_plotRequestMutex.Init() != 0)
	{
		std::cerr << "Fatal error: unable to initialize a mutex" << std::endl;
		exit(-1);
	}
}

ShellMainWin::~ShellMainWin()
{
	m_plotRequestMutex.Lock();

	while (!m_plot2DRequestQueue.empty())
	{
		delete *(m_plot2DRequestQueue.begin());
		m_plot2DRequestQueue.pop_front();
	}

	while (!m_plot3DRequestQueue.empty())
	{
		delete *(m_plot3DRequestQueue.begin());
		m_plot3DRequestQueue.pop_front();
	}
	m_plotRequestMutex.Unlock();
}

void ShellMainWin::startThread()
{
	m_pThread->start();
}

void ShellMainWin::lineEntered()
{
	QString textContents = m_pLineEdit->text();

	if (textContents.length() == 0)
		return;

	m_pLineEdit->setDisabled(true);

	QtIOSystem *pIOSys = (QtIOSystem *)ShellInstance::getInstance()->getIOSystem();
	pIOSys->enterInputLine(textContents.toStdString());
}

void ShellMainWin::requestWriteOutputLine(const std::string &line, bool newline)
{
	QString textContents = QString(line.c_str());

	emit signalWriteOutputLine(textContents, newline);
}

void ShellMainWin::writeOutputLine(const QString &line, bool newline)
{
	QString textContents = line;
	if (newline)
		textContents += QString("\n");
	
	m_pTextEdit->setText(m_pTextEdit->toPlainText() + textContents);
	m_pTextEdit->moveCursor(QTextCursor::End);
}

void ShellMainWin::closeEvent(QCloseEvent *pEvent)
{
	if (m_pThread->isRunning())
		pEvent->ignore();
	else
	{
		// Clean up 2D plots
		{
			std::map<int, Shell2DPlotWindow *>::iterator it;

			for (it = m_2DPlots.begin() ; it != m_2DPlots.end() ; it++)
				delete (*it).second;
			m_2DPlots.clear();
		}
		// Clean up 3D plots
		{
			std::map<int, Shell3DPlotWindow *>::iterator it;

			for (it = m_3DPlots.begin() ; it != m_3DPlots.end() ; it++)
				delete (*it).second;
			m_3DPlots.clear();
		}
		pEvent->accept();
	}
}

void ShellMainWin::waitingForInput()
{
	emit signalWaitingForInput();
}

void ShellMainWin::slotWaitingForInput()
{
	m_pLineEdit->setDisabled(false);
	m_pLineEdit->setFocus();
	m_pLineEdit->setText("");
}

void ShellMainWin::slotForceQuit()
{
	QMessageBox::StandardButton button;

	button = QMessageBox::warning(this, "WARNING", "This will interrupt any computation and will force the program to quit. This should only be used if typing commands is no longer possible. Are you sure you want to quit?",
			              QMessageBox::Yes|QMessageBox::No, QMessageBox::No);

	if (button == QMessageBox::Yes)
	{
		m_pThread->terminate();
		m_pThread->wait(2);
		exit(-1);
	}
}

void ShellMainWin::requestNewPercentage(double pct)
{
	int value = (int)(pct*100.0);

	emit signalProgressBarValue(value);
}

bool ShellMainWin::requestPlot2D(int id, const std::string &title,
			    const std::string &xLabel, const std::string &yLabel,
			    const std::vector<double> &xPoints, const std::vector<double> &yPoints,
			    bool forceShow)
{
	if (xPoints.size() != yPoints.size())
	{
		setErrorString("Number of coordinates in x-direction does not match the number of values in the y-direction");
		return false;
	}

	Shell2DPlotRequest *pRequest = new Shell2DPlotRequest(id, title, xLabel, yLabel, xPoints, yPoints, forceShow);

	m_plotRequestMutex.Lock();
	m_plot2DRequestQueue.push_back(pRequest);
	m_plotRequestMutex.Unlock();

	emit signalNew2DPlotRequest();

	return true;
}

bool ShellMainWin::requestPlot3D(int id, const std::string &title, 
			  const std::string &xLabel, const std::string &yLabel,
			  const std::string &zLabel,
			  double xOff, double yOff, double xDist, double yDist,
			  double xMultiplier, double yMultiplier, double zMultiplier,
			  const std::vector<double> &data, int numX, int numY, bool forceShow)
{
#ifdef SHELLP_CONFIG_OPENGL
	double **pData = new double*[numX];

	for (int i = 0 ; i < numX ; i++)
		pData[i] = new double[numY];

	for (int x = 0 ; x < numX ; x++)
		for (int y = 0 ; y < numY ; y++)
			pData[x][y] = data[x+y*numX]*zMultiplier;

	Shell3DPlotRequest *pRequest = new Shell3DPlotRequest(id, title, xLabel, yLabel, zLabel, 
			                        xOff*xMultiplier, yOff*yMultiplier, xDist*xMultiplier, yDist*yMultiplier, 
						pData, numX, numY, forceShow);

	m_plotRequestMutex.Lock();
	m_plot3DRequestQueue.push_back(pRequest);
	m_plotRequestMutex.Unlock();

	emit signalNew3DPlotRequest();
	return true;
#else
	setErrorString("No OpenGL was enabled at compile time, 3D plotting not supported.");
	return false;
#endif // SHELLP_CONFIG_OPENGL
}

void ShellMainWin::slotNew2DPlotRequest()
{
	Shell2DPlotRequest *pRequest = 0;

	m_plotRequestMutex.Lock();
	if (!m_plot2DRequestQueue.empty())
	{
		pRequest = *(m_plot2DRequestQueue.begin());
		m_plot2DRequestQueue.pop_front();
	}
	m_plotRequestMutex.Unlock();

	if (pRequest == 0)
		return;

	int id = pRequest->getID();

	std::map<int, Shell2DPlotWindow *>::iterator it;
	Shell2DPlotWindow *pPlotWin = 0;
	
	if ((it = m_2DPlots.find(id)) == m_2DPlots.end())
	{
		pPlotWin = new Shell2DPlotWindow();
		m_2DPlots[id] = pPlotWin;
	}
	else
		pPlotWin = (*it).second;

	pPlotWin->processRequest(pRequest);

	//delete pRequest; // is now done in plot window
}

void ShellMainWin::slotNew3DPlotRequest()
{
#ifdef SHELLP_CONFIG_OPENGL
	Shell3DPlotRequest *pRequest = 0;

	m_plotRequestMutex.Lock();
	if (!m_plot3DRequestQueue.empty())
	{
		pRequest = *(m_plot3DRequestQueue.begin());
		m_plot3DRequestQueue.pop_front();
	}
	m_plotRequestMutex.Unlock();

	if (pRequest == 0)
		return;

	int id = pRequest->getID();

	std::map<int, Shell3DPlotWindow *>::iterator it;
	Shell3DPlotWindow *pPlotWin = 0;
	
	if ((it = m_3DPlots.find(id)) == m_3DPlots.end())
	{
		pPlotWin = new Shell3DPlotWindow();
		m_3DPlots[id] = pPlotWin;
	}
	else
		pPlotWin = (*it).second;

	pPlotWin->processRequest(pRequest);

	//delete pRequest; // is now done in plot window
#endif // SHELLP_CONFIG_OPENGL
}

void ShellMainWin::slotInterrupt()
{
	ShellInstance *pInst = ShellInstance::getInstance();
	IOSystem *pIOSys = pInst->getIOSystem();

	pIOSys->writeOutputLine("Requesting interrupt of current calculation");
	pInst->setInterrupted();
}

void ShellMainWin::requestStatusLine(const std::string &line)
{
	QString text = QString(line.c_str());

	emit signalStatusLine(text);
}

void ShellMainWin::slotStatusLine(const QString &line)
{
	if (line.length() == 0)
		setStatusBar(0);
	else
		statusBar()->showMessage(line);
}

} // end namespace
